import logging

from celery import shared_task
from django.core.exceptions import ValidationError

from whatsapp_business_api_is.messages import send_error_message, \
    send_unknown_message, get_next_message, send_next_message
from whatsapp_business_api_is.models import WaUser, OutgoingMessage
from whatsapp_business_api_is.user_msg import msg_factory
from whatsapp_business_api_is.utils import get_start_message, \
    validate_value, \
    run_actions, \
    get_quick_replies_as_flat_list


@shared_task
def async_send_message(user_id, msg, incoming_message_id, reply_message_id):
    logging.info(f"About to send {reply_message_id} to {user_id}")
    user = WaUser.objects.filter(number=user_id).first()
    incoming_message = OutgoingMessage.objects.filter(pk=incoming_message_id).first()

    reply_message = get_next_message(user,
                                     None,
                                     next_message=OutgoingMessage.objects.get(pk=reply_message_id))

    send_next_message(user, msg, incoming_message, reply_message)


@shared_task
def parse_incoming_message(raw_msg):
    reply_message = None
    incoming_message = None
    msg = msg_factory(raw_msg)
    msg_type = msg.type
    logging.debug(f'{msg.__dict__=}')
    ignore_validation = False

    user, _ = WaUser.objects.get_or_create(number=msg.number)

    if user.disable_bot:
        logging.debug(f'Bot is disabled for {user}')
        return
    if msg_type == 'text' and (incoming_message := get_start_message(msg.text)):

        logging.info(f"Start message")

        reply_message = incoming_message.reply
    elif user.state == 'initial':
        if initial_welcome_message := OutgoingMessage.objects.filter(key='initial_welcome_message').first():
            logging.info(f"Unknown message from new user")
            send_next_message(user, None, None, initial_welcome_message)
        else:
            send_unknown_message(user)
        return
    elif msg_type == 'text' and msg.text == '*':
        reply_message = get_next_message(user, None, user.state)
        ignore_validation = True
    elif (msg_type == 'button' and msg.button_payload == "wab_is_do_nothing") or \
            (msg_type == "interactive" and msg.button_reply_id == "wab_is_do_nothing"):
        return
    else:
        current_state = user.state
        logging.info(f"{current_state=}")
        logging.debug(f"{current_state.responses.all()=} {current_state.responses.exists()=}")
        if not current_state.responses.exists():
            if no_waiting_response_message := OutgoingMessage.objects.filter(key='no_waiting_response_message').first():
                logging.info(f"No waiting response")
                send_next_message(user, None, None, no_waiting_response_message)
            send_unknown_message(user)
            return

        button_key = ''
        match current_state.type:

            case 'quick_reply':
                if msg_type == 'interactive':
                    button_key = msg.button_reply_id
                else:
                    button_text = ''
                    if msg_type == 'button':
                        button_text = msg.button_text

                    elif msg_type == 'text':
                        button_text = msg.text

                    for button_key, button_pattern in get_quick_replies_as_flat_list(current_state.quick_reply):
                        if button_pattern == button_text:
                            break
                    else:
                        send_unknown_message(user)
                        return

                logging.debug(f"{button_key=}")
                incoming_message = current_state.responses.filter(key=button_key).first()
            case 'choices':
                msg_text = msg.text if hasattr(msg, 'text') else None
                incoming_message = None
                choice_key_prefix = f"{current_state.key}_resp_"
                logging.debug(f'{choice_key_prefix=}')
                if msg_text:
                    incoming_message = current_state.responses.filter(key__startswith=choice_key_prefix,
                                                                      pattern=msg_text).first()
                if not incoming_message:
                    incoming_message = current_state.responses.filter(key=f"{choice_key_prefix}_default_choice").first()
            case 'text':
                incoming_message = current_state.responses.first()
            case _:
                incoming_message = current_state.responses.first()
        logging.info(f"{incoming_message=}")
        if not incoming_message:
            logging.info(f"No message found")
            send_unknown_message(user)
            return

    try:
        if not ignore_validation:
            validate_value(user, msg, incoming_message)
        run_actions(user, msg, incoming_message)
    except ValidationError as e:
        logging.error(f"{e.message=}")
        send_error_message(user, e)
        return

    if not reply_message:
        reply_message = get_next_message(user, incoming_message)

    send_next_message(user, msg, incoming_message, reply_message)



import logging
from inspect import getmembers

from django.apps import AppConfig
from django.conf import settings
from django.utils.module_loading import import_string

from whatsapp_business_api_is.conf import Conf


class WhatsappBusinessApiIsConfig(AppConfig):
    """
    # General configuration for Whatsapp Business Api Infrastructure

    ## incoming_parser
    ### a function loaded from `Conf.INCOMING_PARSER`.
    The `incoming_parser` receive the raw msg object as received from Whatsapp webhook
    By default it uses `whatsapp_business_api_is.tasks.parse_incoming_message`

    ## set_state
    ### a function to run when the user state is changed

    ## FUNCTIONS
    ### a flat dictionary for all methods of `Functions` class defined in `bot_functions.py` of all registered apps
    Those functions can be used by the messages objects by name.
    Each function name must be unique.

    The signature for all functions must be `user, msg, msg_obj=None, data=None`
    @user - the WaUser object
    @msg - the received user message (instance of `BaseMsg`, depend on the received message)
    @msg_obj - can be either OutgoingMessage or IncomingMessage
    @data - the action's data as defined in the msg_obj

    Example:

    An IncomingMessage object:
    ```
    - model: whatsapp_business_api_is.IncomingMessage
      pk: incoming_example
      fields:
        reply: outgoing_example
        type: user_start
        pattern: "example"
        actions:
          function1:
    ```
    An OutgoingMessage object:
    ```
    - model: whatsapp_business_api_is.IncomingMessage
      pk: outgoing_example
        fields:
          type: text
          text: |
            Hello
          actions:
            function2:
              field: "example"
    ```
    functions objects:
    ```
    #app1/bot_functions.py

    class Functions:
        @staticmethod
        def function1(user, msg, msg_obj=None, data=None):
            pass
    ```
    ```
    #app2/bot_functions.py

    class Functions:
        @staticmethod
        def function2(user, msg, msg_obj=None, data=None):
            pass
    ```

    When a user with the number 1234 will send "example", first we run `function1` with:
    ```
    user=WaUser(number='12345')
    msg=TextMsg(text="example")
    msg_obj=IncomingMessage(pk="incoming_example")
    data=None
    ```
    Then before sending `outgoing_example` we run `function2` with:
    ```
    user=WaUser(number='12345')
    msg=TextMsg(text="example")
    msg_obj=OutgoingMessage(pk="outgoing_example")
    data={ field: "example" }
    ```

    ## VALIDATORS
    ### a flat dictionary for all methods of `Validators` class defined in `bot_validators.py` of all registered apps
    Those functions can be used by the messages objects by name.
    Unlike FUNCTIONS, VALIDATORS are saved with the app label {app}.{validator}

    The signature for all validators must be `user, msg, msg_obj, validation_message=''`
    @user - the WaUser object
    @msg - the received user message (instance of `BaseMsg`, depend on the received message)
    @msg_obj - can be either OutgoingMessage or IncomingMessage
    @validation_message - custom text to be sent to the user in case of validation e

    In case of validation error the method should call `raise ValidationError(validation_message, params={'custom_message': True})`
    The validators are called by `whatsapp_business_api_is.utils.validate_value`.

    If there is no validation message, the bot will send `OutgoingMessage(pk="wrong_format")` message.
    """
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'whatsapp_business_api_is'
    FUNCTIONS = {}
    VALIDATORS = {}
    incoming_parser = None
    set_state = None

    def ready(self):
        try:
            WhatsappBusinessApiIsConfig.incoming_parser = import_string(Conf.INCOMING_PARSER)
            assert callable(WhatsappBusinessApiIsConfig.incoming_parser)
        except AssertionError:
            raise ValueError(f'Conf.INCOMING_PARSER is not a function. Got: {Conf.INCOMING_PARSER}')

        if Conf.SET_STATE:
            try:
                WhatsappBusinessApiIsConfig.set_state = import_string(Conf.SET_STATE)
                assert callable(WhatsappBusinessApiIsConfig.set_state)
            except AssertionError:
                raise ValueError(f'Conf.SET_STATE is not a function. Got: {Conf.SET_STATE}')

        for app in settings.INSTALLED_APPS:

            try:
                validators = import_string(f"{app}.bot_validators.Validators")

                for name, func in getmembers(validators):
                    if name.startswith('_'):
                        continue
                    self.VALIDATORS[f"{app}.{name}"] = func
            except ImportError:
                pass

            try:
                functions = import_string(f"{app}.bot_functions.Functions")

                for name, func in getmembers(functions):
                    if name.startswith('_'):
                        continue
                    if name in self.FUNCTIONS:
                        raise ValueError(f"duplicate key '{name}' found")
                    self.FUNCTIONS[name] = func
            except ImportError:
                pass

        logging.debug("\n\n[Functions]\n  . " + '\n  . '.join(self.FUNCTIONS.keys()))
        logging.debug("\n\n[validators]\n  . " + '\n  . '.join(self.VALIDATORS.keys()))



# Generated by Django 4.0.3 on 2022-12-15 20:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whatsapp_business_api_is', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='wauser',
            name='disable_bot',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='wauser',
            name='failure_count',
            field=models.IntegerField(default=0),
        ),
    ]
